# Copyright (c) 2025, NVIDIA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.  THIS SOFTWARE IS
# PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
# NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import itertools
import os
import sys

from pyeclib import cli
from pyeclib import ec_iface


def add_verify_args(parser):
    parser.add_argument("-q", "--quiet", action="store_true")
    cli.add_instance_args(parser)


def verify_command(args):
    args.ec_type = cli.expand_ec_types(args.ec_type)
    total_failures = 0
    total_corrupt = 0
    data = os.urandom(args.segment_size)
    width = max(len(ec_type) for ec_type in args.ec_type)
    print(f"Using {args.n_data} data + {args.n_parity} parity with "
          f"{args.unavailable} unavailable frags")

    for ec_type in args.ec_type:
        if ec_type not in ec_iface.ALL_EC_TYPES:
            print(f"{ec_type:<{width}} unknown")
            continue
        if ec_type not in ec_iface.VALID_EC_TYPES:
            print(f"{ec_type:<{width}} not available")
            continue
        try:
            instance = ec_iface.ECDriver(
                ec_type=ec_type,
                k=args.n_data,
                m=args.n_parity,
            )
        except ec_iface.ECDriverError:
            print(f"{ec_type:<{width}} could not be instantiated")
            continue
        frags = instance.encode(data)
        combinations, failures, corrupt = check_instance(
            instance, frags, args.unavailable, data)
        total_failures += failures
        total_corrupt += corrupt
        if corrupt:
            print(f"\x1b[91;40m{ec_type:<{width}} {combinations=}, "
                  f"{failures=}, {corrupt=}\x1b[0m")
        elif failures:
            print(f"\x1b[1;91m{ec_type:<{width}} {combinations=}, "
                  f"{failures=}, {corrupt=}\x1b[0m")
        else:
            print(f"{ec_type:<{width}} {combinations=}")

    if total_corrupt:
        return 3
    if total_failures:
        return 1
    return 0


def check_instance(instance, frags, unavailable, data):
    combinations = corrupt = failures = 0
    for to_decode in itertools.combinations(frags, len(frags) - unavailable):
        combinations += 1
        try:
            rebuilt = instance.decode(to_decode)
            if rebuilt != data:
                corrupt += 1
        except ec_iface.ECDriverError:
            failures += 1
            # Might want to log?
    return combinations, failures, corrupt


verify_description = "validate reconstructability of EC schemas"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=verify_description)
    add_verify_args(parser)
    args = parser.parse_args()
    sys.exit(verify_command(args))



# Copyright (c) 2025, NVIDIA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.  THIS SOFTWARE IS
# PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
# NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import sys

from pyeclib import ec_iface


def add_check_args(parser):
    parser.add_argument("-q", "--quiet", action="store_true")
    parser.add_argument("ec_type")


def check_command(args):
    if args.ec_type in ec_iface.VALID_EC_TYPES:
        if not args.quiet:
            print(args.ec_type, 'is available')
        return 0

    if args.ec_type in ec_iface.ALL_EC_TYPES:
        if not args.quiet:
            print(args.ec_type, 'is missing')
        return 1

    if not args.quiet:
        print(args.ec_type, 'is unknown')
    return 2


check_description = "check for backend availability"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=check_description)
    add_check_args(parser)
    args = parser.parse_args()
    sys.exit(check_command(args))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Build a pyeclib wheel that contains precompiled liberasurecode libraries.

The goal is to build manylinux, abi3 wheels that are actually useful.

   - ``manylinux`` ensures installability on a variety of distributions,
     almost regardless of libc version.
   - ``abi3`` ensures compatibility across a variety of python minor versions.
   - "Actually useful" means you can not only import pyeclib, but use it to
     perform encoding/decoding.
   - Where possible, we want to bundle in ISA-L support, too.

You might expect ``auditwheel repair`` to be able to do this for us. However,
that's primarily designed around dynamic *linking* -- and while that's
necessary, it's not sufficient; liberasurecode makes extensive use of dynamic
*loading* as well, and so the dynamically loaded modules need to be included.
"""

import argparse
import base64
import errno
import functools
import hashlib
import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import zipfile


ENV_KEY = ('DYLD_LIBRARY_PATH' if sys.platform == 'darwin'
           else 'LD_LIBRARY_PATH')
if ENV_KEY in os.environ:
    os.environ[ENV_KEY] += ':/usr/lib:/usr/local/lib'
else:
    os.environ[ENV_KEY] = '/usr/lib:/usr/local/lib'


def locate_library(name, missing_ok=False):
    """
    Find a library.

    :param name: The name of the library to find, not including the
                 leading "lib".
    :param missing_ok: If true, return ``None`` when the library cannot be
                       located.
    :raises RuntimeError: If the library cannot be found and ``missing_ok``
                          is ``False``.
    :returns: The full path to the library.
    """
    expr = r'[^\(\)\s]*lib%s\.[^\(\)\s]*' % re.escape(name)
    cmd = ['ld', '-t']
    if sys.platform == 'darwin':
        cmd.extend(['-arch', platform.machine()])
    libpath = os.environ.get(ENV_KEY)
    if libpath:
        for d in libpath.split(':'):
            cmd.extend(['-L', d.rstrip('/')])
    should_read_path = False
    for flg in os.environ.get('LDFLAGS', '').split(' '):
        if should_read_path:
            cmd.append(flg)
            should_read_path = False
        elif flg == '-L':
            cmd.append(flg)
            should_read_path = True
        elif flg.startswith('-L'):
            cmd.append(flg)
    cmd.extend(['-o', os.devnull, '-l%s' % name])
    p = subprocess.run(cmd, capture_output=True, text=True)
    # Note that we don't expect the command to exit cleanly; we just want
    # to see what got loaded
    res = re.search(expr, p.stdout)
    if res:
        return os.path.realpath(res.group(0))

    if missing_ok:
        return None

    print(f'{" ".join(cmd)!r} failed with status {p.returncode}:')
    print(f'\x1b[31m{p.stderr}\x1b[0m', end='')
    raise RuntimeError(f'Failed to locate {name}')


def build_wheel(src_dir):
    """
    Build the base wheel, returning the path to the wheel.

    Caller is responsible for cleaning up the tempdir.
    """
    tmp = tempfile.mkdtemp()
    try:
        subprocess.check_call([
            sys.executable, 'setup.py',
            'bdist_wheel', '-d', tmp, '--py-limited-api=cp310',
        ], cwd=src_dir)
        files = os.listdir(tmp)
        assert len(files) == 1, files
        return os.path.join(tmp, files[0])
    except Exception:
        shutil.rmtree(tmp)
        raise


def repack_wheel(whl, so_suffix, out_whl=None, require_isal=False):
    """
    Repack a wheel to bundle in liberasurecode libraries.

    This unpacks the wheel, copies all the supporting libraries to the
    unpacked wheel, adjusts rpath etc for the libraries, rebuilds the
    dist-info to include the libraries, and rebuilds the wheel.
    """
    if out_whl is None:
        out_whl = whl

    tmp = tempfile.mkdtemp()
    try:
        # unpack wheel
        zf = zipfile.ZipFile(whl, 'r')
        zf.extractall(tmp)

        relocate_libs(tmp, so_suffix, require_isal)
        rebuild_dist_info_record(tmp)
        build_zip(tmp, out_whl)
    finally:
        shutil.rmtree(tmp)


def relocate_libs(tmp, so_suffix, require_isal=False):
    """
    Bundle libraries into a unpacked-wheel tree.

    :param tmp: the temp dir containing the tree for the unzipped wheel
    :param so_suffix: the LIBERASURECODE_SO_SUFFIX used to build
                      liberasurecode; this should be used to avoid
                      interfering with system libraries
    """
    lib_dir = 'pyeclib.libs'
    all_libs = [os.path.join(tmp, lib)
                for lib in os.listdir(tmp) if '.so' in lib]
    for lib in all_libs:  # NB: pypy builds may create multiple .so's
        update_rpath(lib, '/' + lib_dir)

    inject = functools.partial(
        inject_lib,
        tmp,
        so_suffix=so_suffix,
        lib_dir=lib_dir,
        all_libs=all_libs,
    )

    # Be sure to move liberasurecode first, so we can fix its links to
    # the others
    relocated_libec = inject(locate_library('erasurecode'))
    # Since liberasurecode links against other included libraries,
    # need to update rpath
    update_rpath(relocated_libec)

    # These guys all stand on their own, so don't need the rpath update
    inject(locate_library('nullcode'))
    inject(locate_library('Xorcode'))
    builtin_rs_vand = inject(locate_library('erasurecode_rs_vand'))
    maybe_add_needed(relocated_libec, os.path.basename(builtin_rs_vand))

    # Nobody actually links against this, but we want it anyway if available
    isal = locate_library('isal', missing_ok=not require_isal)
    if isal:
        relocated_isal = inject(isal)
        maybe_add_needed(relocated_libec, os.path.basename(relocated_isal))


def update_rpath(lib, rpath_suffix=''):
    if sys.platform == 'darwin':
        subprocess.check_call([
            'install_name_tool',
            '-add_rpath', '@loader_path' + rpath_suffix,
            lib,
        ])
    else:
        subprocess.check_call([
            'patchelf', '--set-rpath', '$ORIGIN' + rpath_suffix, lib])


def maybe_add_needed(lib, needed):
    if sys.platform == 'linux' and platform.libc_ver()[0] != 'glibc':
        # We need to do this for musl; it doesn't seem to respect RUNPATH
        # for dynamic loading, just dynamic linking
        subprocess.check_call(['patchelf', '--add-needed', needed, lib])


def inject_lib(
    whl_dir,
    src_lib,
    so_suffix='-pyeclib',
    lib_dir='pyeclib.libs',
    all_libs=None,
):
    try:
        os.mkdir(os.path.join(whl_dir, lib_dir))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    if sys.platform == 'darwin':
        old_lib = src_lib
        name = os.path.basename(old_lib).split('.', 1)[0]
        new_lib = name + so_suffix + '.dylib'
    else:
        name, _, version = os.path.basename(src_lib).partition('.so')
        major = '.'.join(version.split('.', 2)[:2])
        old_lib = name + '.so' + major
        new_lib = name + so_suffix + '.so' + major

    print('Injecting ' + new_lib)
    relocated = os.path.join(whl_dir, lib_dir, new_lib)
    shutil.copy2(src_lib, relocated)
    if sys.platform == 'darwin':
        subprocess.check_call([
            'install_name_tool', '-id', new_lib, relocated])
    else:
        subprocess.check_call(['patchelf', '--set-soname', new_lib, relocated])

    if all_libs:
        # Fix linkage in the libs already moved -- this is mainly an issue for
        # liberasurecode.so. Jerasure *would* need it for GF-Complete, but it
        # seems unlikely we'd be able to include those any time soon
        if sys.platform == 'darwin':
            for lib in all_libs:
                subprocess.check_call([
                    'install_name_tool',
                    '-change', old_lib,
                    '@rpath/' + new_lib,
                    lib,
                ])
        else:
            subprocess.check_call([
                'patchelf', '--replace-needed', old_lib, new_lib] + all_libs)
        all_libs.append(relocated)

    return relocated


def rebuild_dist_info_record(tmp):
    """
    Update the dist-info RECORD information.

    There are likely new files, and pre-existing files may have changed;
    rebuild the whole thing.

    See https://packaging.python.org/en/latest/specifications/
    recording-installed-packages/#the-record-file for more info.
    """
    tmp = tmp.rstrip('/') + '/'
    dist_info_dir = [d for d in os.listdir(tmp) if d.endswith('.dist-info')]
    assert len(dist_info_dir) == 1, dist_info_dir
    record_file = os.path.join(tmp, dist_info_dir[0], 'RECORD')
    with open(record_file, 'w') as fp:
        for dir_path, _, files in os.walk(tmp):
            for file in files:
                file = os.path.join(dir_path, file)
                if file == record_file:
                    fp.write(file[len(tmp):] + ',,\n')
                    continue
                hsh, sz = sha256(file)
                fp.write('%s,sha256=%s,%d\n' % (file[len(tmp):], hsh, sz))


def sha256(file):
    hasher = hashlib.sha256()
    sz = 0
    with open(file, 'rb') as fp:
        for chunk in iter(lambda: fp.read(128 * 1024), b''):
            hasher.update(chunk)
            sz += len(chunk)
    hsh = base64.urlsafe_b64encode(hasher.digest())
    return hsh.decode('ascii').strip('='), sz


def build_zip(tmp, out_file):
    """
    Zip up all files in a tree, with archive names relative to the root.
    """
    tmp = tmp.rstrip('/') + '/'
    with zipfile.ZipFile(out_file, 'w') as zf:
        for dir_path, _, files in os.walk(tmp):
            for file in files:
                file = os.path.join(dir_path, file)
                zf.write(file, file[len(tmp):])


def repair_wheel(whl):
    """
    Run ``auditwheel repair`` to ensure appropriate platform tags.
    """
    if sys.platform == 'darwin':
        return whl  # auditwheel only works on linux
    whl_dir = os.path.dirname(whl)
    subprocess.check_call(['auditwheel', 'repair', whl, '-w', whl_dir])
    wheels = [f for f in os.listdir(whl_dir) if f != os.path.basename(whl)]
    assert len(wheels) == 1, wheels
    return os.path.join(whl_dir, wheels[0])


def fix_ownership(whl):
    uid = int(os.environ.get('UID', '1000'))
    gid = int(os.environ.get('GID', uid))
    os.chown(whl, uid, gid)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('src_dir')
    parser.add_argument('-w', '--wheel-dir', default='.')
    parser.add_argument('-s', '--so-suffix', default='-pyeclib')
    parser.add_argument('-r', '--repair', action='store_true')
    parser.add_argument('--require-isal', action='store_true')
    args = parser.parse_args()
    whl = build_wheel(args.src_dir)
    whl_dir = os.path.dirname(whl)
    try:
        repack_wheel(whl, args.so_suffix, require_isal=args.require_isal)
        if args.repair:
            whl = repair_wheel(whl)
        output_whl = os.path.join(
            args.wheel_dir, os.path.basename(whl))
        shutil.move(whl, output_whl)
        if os.geteuid() == 0:
            # high likelihood of running in a docker container or something
            fix_ownership(output_whl)
    finally:
        shutil.rmtree(whl_dir)


if __name__ == '__main__':
    main()



# Copyright (c) 2025, NVIDIA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.  THIS SOFTWARE IS
# PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
# NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pyeclib import ec_iface


_type_abbreviations = {
    # name -> prefix
    "all": "",
    "isa-l": "isa_l_",
    "isa_l": "isa_l_",
    "isal": "isa_l_",
    "jerasure": "jerasure_",
    "flat-xor": "flat_xor_",
    "flat_xor": "flat_xor_",
    "flatxor": "flat_xor_",
    "xor": "flat_xor_",
}


def expand_ec_types(user_types):
    result = set(user_types or ["all"])
    for abbrev, prefix in _type_abbreviations.items():
        if abbrev in result:
            result.remove(abbrev)
            result.update(
                ec_type
                for ec_type in ec_iface.ALL_EC_TYPES
                if ec_type.startswith(prefix)
            )
    return sorted(result)


def add_instance_args(parser, default_segment_size=1024):
    """
    Add arguments to ``parser`` for instance instantiation.
    """
    parser.add_argument(
        "--ec-type",
        action="append",
        type=str,
    )
    parser.add_argument(
        "--n-data",
        "--ndata",
        "-k",
        metavar="K",
        type=int,
        default=10,
    )
    parser.add_argument(
        "--n-parity",
        "--nparity",
        "-m",
        metavar="M",
        type=int,
        default=5,
    )
    parser.add_argument(
        "--unavailable",
        "-u",
        metavar="N",
        type=int,
        default=2,
    )
    parser.add_argument(
        "--segment-size",
        "-s",
        metavar="BYTES",
        type=int,
        default=default_segment_size,
    )



# Copyright (c) 2025, NVIDIA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.  THIS SOFTWARE IS
# PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
# NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import io
import platform
import re
import unittest
from unittest import mock

from pyeclib.cli.__main__ import main
from pyeclib import ec_iface


class TestVersion(unittest.TestCase):
    def _test_version(self, args):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                self.assertRaises(SystemExit) as caught:
            main(args)
        self.assertTrue(stdout.getvalue().endswith("\n"))
        line_parts = [
            line.split(" ")
            for line in stdout.getvalue()[:-1].split("\n")
        ]
        self.assertEqual([prog for prog, vers in line_parts], [
            'pyeclib',
            'liberasurecode',
            platform.python_implementation(),
        ])
        re_version = re.compile(r"\d+\.\d+\.\d+")
        self.assertEqual(
            [bool(re_version.match(vers)) for prog, vers in line_parts],
            [True] * 3)
        self.assertEqual(caught.exception.code, None)

    def test_subcommand(self):
        self._test_version(["version"])

    def test_opt(self):
        self._test_version(["-V"])


class TestList(unittest.TestCase):
    def test_list(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                self.assertRaises(SystemExit) as caught:
            main(["list"])
        self.assertTrue(stdout.getvalue().endswith("\n"))
        line_parts = [
            line.split()
            for line in stdout.getvalue()[:-1].split("\n")
        ]
        self.assertEqual(
            [backend for backend, status in line_parts],
            sorted(ec_iface.ALL_EC_TYPES))
        self.assertEqual(
            {status for backend, status in line_parts},
            {"available", "missing"})
        self.assertEqual(caught.exception.code, 0)

    def test_list_one(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                self.assertRaises(SystemExit) as caught:
            main(["list", "liberasurecode_rs_vand"])
        self.assertTrue(stdout.getvalue().endswith("\n"))
        line_parts = [
            line.split()
            for line in stdout.getvalue()[:-1].split("\n")
        ]
        self.assertEqual(line_parts, [["liberasurecode_rs_vand", "available"]])
        self.assertEqual(caught.exception.code, 0)

    def test_list_one_unknown(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                self.assertRaises(SystemExit) as caught:
            main(["list", "missing-backend"])
        self.assertTrue(stdout.getvalue().endswith("\n"))
        line_parts = [
            line.split()
            for line in stdout.getvalue()[:-1].split("\n")
        ]
        self.assertEqual(line_parts, [["missing-backend", "unknown"]])
        self.assertEqual(caught.exception.code, 1)

    def test_list_multiple_mixed_status(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                self.assertRaises(SystemExit) as caught:
            main(["list", "missing-backend", "liberasurecode_rs_vand"])
        self.assertTrue(stdout.getvalue().endswith("\n"))
        line_parts = [
            line.split()
            for line in stdout.getvalue()[:-1].split("\n")
        ]
        self.assertEqual(line_parts, [
            ["liberasurecode_rs_vand", "available"],
            ["missing-backend", "unknown"],
        ])
        self.assertEqual(caught.exception.code, 0)

    def test_list_abbreviation(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                self.assertRaises(SystemExit) as caught:
            main(["list", "isal"])
        self.assertTrue(stdout.getvalue().endswith("\n"))
        line_parts = [
            line.split()
            for line in stdout.getvalue()[:-1].split("\n")
        ]
        self.assertEqual(
            [backend for backend, status in line_parts],
            sorted(backend for backend in ec_iface.ALL_EC_TYPES
                   if backend.startswith("isa_l_")))
        found_statuses = {status for backend, status in line_parts}
        if ec_iface.get_liberasurecode_version() >= 0x1_07_00:
            self.assertEqual(len(found_statuses), 1, found_statuses)
        for status in found_statuses:
            self.assertIn(status, {"available", "missing"})
        self.assertEqual(caught.exception.code,
                         0 if "available" in found_statuses else 1)

    def test_list_available(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                self.assertRaises(SystemExit) as caught:
            main(["list", "--available"])
        self.assertTrue(stdout.getvalue().endswith("\n"))
        self.assertEqual(
            stdout.getvalue()[:-1].split("\n"),
            sorted(ec_iface.VALID_EC_TYPES))
        self.assertEqual(caught.exception.code, 0)

    def test_list_available_abbreviation(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                self.assertRaises(SystemExit) as caught:
            main(["list", "--available", "flatxor"])
        self.assertTrue(stdout.getvalue().endswith("\n"))
        self.assertEqual(stdout.getvalue()[:-1].split("\n"), [
            "flat_xor_hd_3",
            "flat_xor_hd_4",
        ])
        self.assertEqual(caught.exception.code, 0)


class TestCheck(unittest.TestCase):
    def test_check_backend_required(self):
        with mock.patch("sys.stderr", new=io.StringIO()) as stderr, \
                self.assertRaises(SystemExit) as caught:
            main(["check"])
        self.assertIn("the following arguments are required: ec_type",
                      stderr.getvalue())
        self.assertEqual(caught.exception.code, 2)

    def test_check_available(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                mock.patch("sys.stderr", new=io.StringIO()) as stderr, \
                self.assertRaises(SystemExit) as caught:
            main(["check", "liberasurecode_rs_vand"])
        self.assertEqual("liberasurecode_rs_vand is available\n",
                         stdout.getvalue())
        self.assertEqual("", stderr.getvalue())
        self.assertEqual(caught.exception.code, 0)

    def test_check_missing(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                mock.patch("sys.stderr", new=io.StringIO()) as stderr, \
                mock.patch("pyeclib.ec_iface.VALID_EC_TYPES", []), \
                self.assertRaises(SystemExit) as caught:
            main(["check", "liberasurecode_rs_vand"])
        self.assertEqual("liberasurecode_rs_vand is missing\n",
                         stdout.getvalue())
        self.assertEqual("", stderr.getvalue())
        self.assertEqual(caught.exception.code, 1)

    def test_check_unknown(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                mock.patch("sys.stderr", new=io.StringIO()) as stderr, \
                mock.patch("pyeclib.ec_iface.VALID_EC_TYPES", []), \
                self.assertRaises(SystemExit) as caught:
            main(["check", "unknown-backend"])
        self.assertEqual("unknown-backend is unknown\n",
                         stdout.getvalue())
        self.assertEqual("", stderr.getvalue())
        self.assertEqual(caught.exception.code, 2)

    def test_check_quiet_available(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                mock.patch("sys.stderr", new=io.StringIO()) as stderr, \
                self.assertRaises(SystemExit) as caught:
            main(["check", "-q", "liberasurecode_rs_vand"])
        self.assertEqual("", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())
        self.assertEqual(caught.exception.code, 0)

    def test_check_quiet_missing(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                mock.patch("sys.stderr", new=io.StringIO()) as stderr, \
                mock.patch("pyeclib.ec_iface.VALID_EC_TYPES", []), \
                self.assertRaises(SystemExit) as caught:
            main(["check", "--quiet", "liberasurecode_rs_vand"])
        self.assertEqual("", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())
        self.assertEqual(caught.exception.code, 1)

    def test_check_quiet_unknown(self):
        with mock.patch("sys.stdout", new=io.StringIO()) as stdout, \
                mock.patch("sys.stderr", new=io.StringIO()) as stderr, \
                mock.patch("pyeclib.ec_iface.VALID_EC_TYPES", []), \
                self.assertRaises(SystemExit) as caught:
            main(["check", "-q", "unknown-backend"])
        self.assertEqual("", stdout.getvalue())
        self.assertEqual("", stderr.getvalue())
        self.assertEqual(caught.exception.code, 2)



# Copyright (c) 2025, NVIDIA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.  THIS SOFTWARE IS
# PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
# NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import sys

from pyeclib.cli import bench
from pyeclib.cli import check
from pyeclib.cli import list as list_cli
from pyeclib.cli import verify
from pyeclib.cli import version


def main(args=None):
    parser = argparse.ArgumentParser(
        prog="pyeclib-backend",
        description="tool to get various erasure coding information",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="store_const",
        dest="func",
        const=version.version_command,
        help=version.version_description,
    )
    subparsers = parser.add_subparsers()

    version_parser = subparsers.add_parser(
        "version", help=version.version_description)
    version_parser.set_defaults(func=version.version_command)

    list_parser = subparsers.add_parser(
        "list", help=list_cli.list_description)
    list_parser.set_defaults(func=list_cli.list_command)
    list_cli.add_list_args(list_parser)

    check_parser = subparsers.add_parser(
        "check", help=check.check_description)
    check_parser.set_defaults(func=check.check_command)
    check.add_check_args(check_parser)

    verify_parser = subparsers.add_parser(
        "verify", help=verify.verify_description)
    verify_parser.set_defaults(func=verify.verify_command)
    verify.add_verify_args(verify_parser)

    bench_parser = subparsers.add_parser(
        "bench", help=bench.bench_description)
    bench_parser.set_defaults(func=bench.bench_command)
    bench.add_bench_args(bench_parser)

    parsed_args = parser.parse_args(args)
    if parsed_args.func is None:
        parser.error(
            f"the following arguments are required: "
            f"{{{','.join(subparsers.choices)}}}"
        )
    sys.exit(parsed_args.func(parsed_args))


if __name__ == "__main__":
    main()



# Copyright (c) 2025, NVIDIA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.  THIS SOFTWARE IS
# PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
# NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import sys

from pyeclib import cli
from pyeclib import ec_iface


def add_list_args(parser):
    parser.add_argument("-a", "--available", action="store_true",
                        help="display only available backends")
    parser.add_argument("ec_type", nargs="*", type=str,
                        help="display these backends (default: all)")


def list_command(args):
    args.ec_type = cli.expand_ec_types(args.ec_type)

    width = max(len(backend) for backend in args.ec_type)
    found = 0
    for backend in args.ec_type:
        if args.available:
            if backend in ec_iface.VALID_EC_TYPES:
                print(backend)
                found += 1
        else:
            if backend not in ec_iface.ALL_EC_TYPES:
                print(f"{backend:<{width}} unknown")
            elif backend in ec_iface.VALID_EC_TYPES:
                print(f"{backend:<{width}} available")
                found += 1
            else:
                print(f"{backend:<{width}} missing")
    return 0 if found else 1


list_description = "list availability of EC backends"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=list_description)
    add_list_args(parser)
    args = parser.parse_args()
    sys.exit(list_command(args))



# Copyright (c) 2025, NVIDIA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.  THIS SOFTWARE IS
# PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
# NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import random
import time

from pyeclib import cli
from pyeclib import ec_iface


def add_bench_args(parser):
    parser.add_argument("-e", "--encode", action="store_true")
    parser.add_argument("-d", "--decode", action="store_true")
    cli.add_instance_args(parser, default_segment_size=2**20)
    parser.add_argument("--iterations", "-i", type=int, default=200)


def bench_command(args):
    args.ec_type = cli.expand_ec_types(args.ec_type)
    data = os.urandom(args.segment_size + args.iterations)
    width = max(len(ec_type) for ec_type in args.ec_type)
    print(f"Using {args.n_data} data + {args.n_parity} parity with "
          f"{args.unavailable} unavailable frags")

    for ec_type in args.ec_type:
        if ec_type not in ec_iface.ALL_EC_TYPES:
            print(f"{ec_type:<{width}} unknown")
            continue
        if ec_type not in ec_iface.VALID_EC_TYPES:
            print(f"{ec_type:<{width}} not available")
            continue
        try:
            instance = ec_iface.ECDriver(
                ec_type=ec_type,
                k=args.n_data,
                m=args.n_parity,
            )
        except ec_iface.ECDriverError:
            print(f"{ec_type:<{width}} could not be instantiated")
            continue
        frags = instance.encode(data[:args.segment_size])

        if args.encode or not args.decode:
            start = time.time()
            for i in range(args.iterations):
                _ = instance.encode(data[i:i + args.segment_size])
            dt = time.time() - start
            mb_encoded = args.iterations * args.segment_size / (2 ** 20)
            print(f'{ec_type} (encode): {mb_encoded / dt:.1f}MB/s')

        if args.decode or not args.encode:
            start = time.time()
            for i in range(args.iterations):
                data_frags = random.sample(
                    frags[:args.n_data],
                    args.n_data - args.unavailable,
                )
                if ec_type.startswith('flat_xor'):
                    # The math is actually more complicated than this, but ...
                    parity_frags = frags[args.n_data:]
                else:
                    parity_frags = random.sample(
                        frags[args.n_data:],
                        args.unavailable,
                    )
                _ = instance.decode(data_frags + parity_frags)
            dt = time.time() - start
            mb_encoded = args.iterations * args.segment_size / (2 ** 20)
            print(f'{ec_type} (decode): {mb_encoded / dt:.1f}MB/s')


bench_description = "benchmark EC schemas"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=bench_description)
    add_bench_args(parser)
    args = parser.parse_args()
    bench_command(args)
